/*
 * redux-sam.js v1.2.0
 * (c) 2018-2019 Jesse Feng
 * Released under the MIT License.
 */
import isFunction from 'celia/isFunction';
import isString from 'celia/isString';
import isObject from 'celia/isObject';
import forOwn from 'celia/forOwn';
import { e as error, s as setNestedState, a as assert, u as unifyObjectStyle, g as getNestedState, r as removeNestedState } from './utils-7fa96983.js';
import isPromiseLike from 'celia/isPromiseLike';
import append from 'celia/_append';

/**
 *  挂载订阅函数
 *
 * @param {Function} fn
 * @param {Array} subs
 */
function genericSubscribe(fn, subs) {
  if (subs.indexOf(fn) === -1) {
    append(subs, fn);
  }
  return function () {
    var i = subs.indexOf(fn);
    if (i > -1) {
      subs.splice(i, 1);
    }
  };
}

/**
 * 重置函数
 * @param {Sam} sam
 * @param {Object} newModule
 */
function resetSam(sam, newModule) {
  sam._actions = Object.create(null);
  sam._mutations = Object.create(null);
  sam._contextNamespaceMap = Object.create(null);

  // 初始化模块
  installModule(sam, [], newModule, true);
}

/**
 * 构建函数中需要的函数
 *
 * @param {Sam} sam
 * @param {String} namespace
 * @param {Array} path
 */
function makeLocalContext(sam, namespace, path) {
  var hasNamespace = namespace !== '';

  var local = {
    dispatch: function dispatch(_type, _payload, _options) {
      var args = unifyObjectStyle(_type, _payload, _options);
      var payload = args.payload;
      var options = args.options;
      var type = args.type;

      if (hasNamespace && !options.root) {
        type = namespace + type;
        if (process.env.NODE_ENV !== 'production' && !sam._actions[type]) {
          error(("unknown local action type: " + (args.type) + ", global type: " + type));
          return;
        }
      }
      return sam.dispatch(type, payload, options);
    },

    commit: function commit(_type, _payload, _options) {
      if (hasNamespace) {
        var args = unifyObjectStyle(_type, _payload, _options);
        var type = args.type;
        var payload = args.payload;
        var options = args.options;

        if (!options.root) {
          type = namespace + type;
          if (process.env.NODE_ENV !== 'production' && !sam._mutations[type]) {
            error(("unknown local mutation type: " + (args.type) + ", global type: " + type));
            return;
          }
        }
        return sam.commit(type, payload, options);
      } else {
        return sam.commit(_type, _payload, _options);
      }
    },

    getState: function getState() {
      return getNestedState(sam, path);
    },

    setState: function setState(ret) {
      isObject(ret) && setNestedState(sam, path, ret);
      return ret;
    }
  };

  return local;
}

/**
 * 安装模块
 *
 * @param {Sam} sam
 * @param {Array|String} path
 * @param {Object} rawModule
 * @param {Boolean} hot
 */
function installModule(sam, path, rawModule, hot) {
  var isRoot = !path.length;

  if (process.env.NODE_ENV !== 'production') {
    assertRawModule(path, rawModule);
  }

  var rawState = rawModule.state;
  var state = (isFunction(rawState) ? rawState() : rawState) || Object.create(null);
  var namespace = path.reduce(function (namespace, key) { return namespace + (rawModule.namespaced ? key + '/' : ''); }, '');
  var local = makeLocalContext(sam, namespace, path);

  // 判断不是顶层模块
  if (!isRoot) {
    // 分类指定命名空间的模块
    if (rawModule.namespaced) {
      if (process.env.NODE_ENV !== 'production' && sam._contextNamespaceMap[namespace]) {
        error(("duplicate namespace " + namespace + " for the namespaced module " + (path.join('/'))));
      }
      sam._contextNamespaceMap[namespace] = local;
    }

    // 挂载子模块到跟模块下
    if (!hot) {
      setNestedState(sam, path, state);
    }
  } else {
    sam._state = state;
  }

  forOwn(rawModule.mutations, function (mutation, key) {
    var namespacedType = namespace + key;
    registerMutation(sam, namespacedType, mutation, local);
  });

  forOwn(rawModule.actions, function (action, key) {
    var type = action.root ? key : namespace + key;
    var handler = action.handler || action;
    registerAction(sam, type, handler, local);
  });

  forOwn(rawModule.modules, function (child, key) {
    installModule(sam, path.concat(key), child, hot);
  });
}

/**
 * 注册mutation
 * @param {Sam} sam
 * @param {String} type
 * @param {Function} handler
 * @param {Object} param3
 */
function registerMutation(sam, type, handler, ref) {
  var setState = ref.setState;
  var getState = ref.getState;

  var entry = sam._mutations[type] || (sam._mutations[type] = []);
  handler = handler.bind(sam);

  entry.push(function (payload) { return setState(handler(getState(), payload)); });
}

/**
 * 注册action
 * @param {Sam} sam
 * @param {String} type
 * @param {Function} handler
 * @param {Object} param3
 */
function registerAction(sam, type, handler, ref) {
  var dispatch = ref.dispatch;
  var commit = ref.commit;
  var getState = ref.getState;

  var entry = sam._actions[type] || (sam._actions[type] = []);
  handler = handler.bind(sam);

  entry.push(function (payload) {
    var res = handler({
      dispatch: dispatch,
      commit: commit,
      state: getState(),
      rootState: sam.state
    }, payload);
    if (!isPromiseLike(res)) {
      res = Promise.resolve(res);
    }
    return res;
  });
}

var assertTypes = {
  mutations: {
    assert: function (value) { return isFunction(value); },
    expected: 'function'
  },
  actions: {
    assert: function (value) { return isFunction(value) ||
      (isObject(value) && isFunction(value.handler)); },
    expected: 'function or object with "handler" function'
  }
};

function makeAssertionMessage(path, key, type, value, expected) {
  var buf = key + " should be " + expected + " but \"" + key + "." + type + "\"";
  if (path.length > 0) {
    buf += " in module \"" + (path.join('.')) + "\"";
  }
  buf += " is " + (JSON.stringify(value)) + ".";
  return buf;
}

function assertRawModule(path, rawModule) {
  forOwn(assertTypes, function (assertOptions, key) {
    // mutations 或者 actions
    var obj = rawModule[key];

    forOwn(obj, function (value, type) {
      assert(
        assertOptions.assert(value),
        makeAssertionMessage(path, key, type, value, assertOptions.expected)
      );
    });
  });
}

/**
 * 创建一个 redux-sam 实例用于解析state、mutations、actions和modules
 *
 * const sam = new Sam({
 *   state: { ... },
 *   mutations: { ... },
 *   actions: { ... }
 *   modules: { ... }
 * });
 *
 * const store = Redux.createStore(reducer(sam), sam.state, applyMiddleware(middleware(sam)));
 *
 */
var Sam = function Sam(options) {
  if ( options === void 0 ) options = {};

  var plugins = options.plugins; if ( plugins === void 0 ) plugins = [];
  var sam = this;

  // 内部状态
  sam._rawRootModule = options;
  sam._state = Object.create(null);
  sam._actions = Object.create(null);
  sam._actionSubscribers = [];
  sam._mutations = Object.create(null);
  sam._subscribers = [];
  sam._contextNamespaceMap = Object.create(null);

  // 初始化state和子模块
  installModule(sam, [], options);

  // 执行插件
  plugins.forEach(function (plugin) { return isFunction(plugin) && plugin(sam); });
};

var prototypeAccessors = { state: { configurable: true } };

prototypeAccessors.state.get = function () {
  return this._state;
};

prototypeAccessors.state.set = function (v) {
  if (process.env.NODE_ENV !== 'production') {
    assert(false, "use sam.replaceState() to explicit replace redux-sam state.");
  }
};

/**
 * 订阅追踪state
 * @param {Function} fn
 */
Sam.prototype.subscribe = function subscribe (fn) {
  return genericSubscribe(fn, this._subscribers);
};

/**
 * 可以订阅action触发前后的公共处理逻辑
 * @param {Function} fn
 */
Sam.prototype.subscribeAction = function subscribeAction (fn) {
  return genericSubscribe(
    isFunction(fn) ? { before: fn } : fn,
    this._actionSubscribers
  );
};

/**
 * 替换state，未来可能会有一些其它逻辑处理
 * @param {Object} state
 */
Sam.prototype.replaceState = function replaceState (state) {
  this._state = state;
};

/**
 * 动态注册state、mutations、actions和modules，可用服务端渲染
 *
 * @param {String|Array} path
 * @param {Object} rawModule
 * @param {Object} options
 */
Sam.prototype.registerModule = function registerModule (path, rawModule, options) {
    if ( options === void 0 ) options = {};

  if (isString(path)) {
    path = [path];
  }

  if (process.env.NODE_ENV !== 'production') {
    assert(Array.isArray(path), "module path must be a string or an Array.");
    assert(path.length > 0, 'cannot register the root module by using registerModule.');
  }

  installModule(
    this,
    path,
    rawModule,
    options.preserveState
  );
};

/**
 * 动态移除已加载的模块
 *
 * @param {String|Array} path
 */
Sam.prototype.unregisterModule = function unregisterModule (path) {
  if (isString(path)) {
    path = [path];
  }

  if (process.env.NODE_ENV !== 'production') {
    assert(Array.isArray(path), "module path must be a string or an Array.");
  }

  removeNestedState(this, path);

  resetSam(this, this._rawRootModule);
};

/**
 * 热更新时调用
 * @param {Object} newOptions
 */
Sam.prototype.hotUpdate = function hotUpdate (newOptions) {
  resetSam(this, newOptions);
};

Object.defineProperties( Sam.prototype, prototypeAccessors );

export default Sam;
