/*
 * redux-sam.js v1.2.0
 * (c) 2018-2019 Jesse Feng
 * Released under the MIT License.
 */
import isFunction from 'celia/isFunction';
import 'celia/isString';
import 'celia/isObject';
import 'celia/forOwn';
import { n as normalizeNamespace, b as normalizeMap, c as getContextByNamespace } from './utils-7fa96983.js';

/**
 * 映射 mutations 组件的实例方法
 *
 * constructor(props) {
 *   super(props);
 *
 *   mapMutations(this, ['increment', 'decrement']);
 * }
 *
 * @param {Sam} sam
 */
function mapMutations(sam) {
  return normalizeNamespace(function (component, mutations, namespace) {
    normalizeMap(mutations, function (val, key) {
      component[key] = function mappedMutation() {
        var args = [], len = arguments.length;
        while ( len-- ) args[ len ] = arguments[ len ];

        var commit = sam.commit;
        if (namespace) {
          var ctx = getContextByNamespace(sam, 'mapMutations', namespace);
          if (!ctx) {
            return;
          }
          commit = ctx.commit;
        }
        return isFunction(val)
          ? val.apply(component, [commit ].concat( args))
          : commit.apply(sam, [val ].concat( args));
      };
    });
    return component;
  });
}
/**
 * 映射 actions 组件的实例方法
 *
 * constructor(props) {
 *   super(props);
 *
 *   mapMutations(this, ['increment', 'decrement', 'incrementIfOdd', 'incrementAsync']);
 * }
 *
 * @param {Sam} sam
 */
function mapActions(sam) {
  return normalizeNamespace(function (component, actions, namespace) {
    normalizeMap(actions, function (val, key) {
      component[key] = function mappedMutation() {
        var args = [], len = arguments.length;
        while ( len-- ) args[ len ] = arguments[ len ];

        var dispatch = sam.dispatch;
        if (namespace) {
          var ctx = getContextByNamespace(sam, 'mapActions', namespace);
          if (!ctx) {
            return;
          }
          dispatch = ctx.dispatch;
        }
        return isFunction(val)
          ? val.apply(component, [dispatch ].concat( args))
          : dispatch.apply(sam, [val ].concat( args));
      };
    });
    return component;
  });
}
/**
 * 创建公用函数
 * @param {Object} sam
 */
function createHelpers (sam) {
  return {
    mapMutations: mapMutations(sam),
    mapActions: mapActions(sam)
  };
}

export default createHelpers;
