/*
 * redux-sam.js v1.2.0
 * (c) 2018-2019 Jesse Feng
 * Released under the MIT License.
 */
'use strict';

require('celia/isString');
require('celia/isObject');
require('celia/forOwn');
var utils = require('./utils-fc5ba3db.js');

/**
 * 触发action订阅
 * @param {Array} actionSubscribers 订阅回调函数集合
 * @param {String} method before或者after方法名
 * @param {Object} action 包括type和payload还有options
 * @param {Object} state 顶层state或者嵌套子state
 */
function triggerActionSubscribers(actionSubscribers, method, action, state) {
  try {
    actionSubscribers
      .forEach(function (sub) {
        var fn = sub[method];
        fn && fn(action, state);
      });
  } catch (e) {
    if (process.env.NODE_ENV !== 'production') {
      utils.error(("error in " + method + " action subscribers: "), e);
    }
  }
}

/**
 * redux中间件
 * @param {Sam} sam
 */
function middleware(sam) {
  return function (ref) {
    var dispatch = ref.dispatch;

    // 挂载 commit 方法 至 sam 对象上，用于触发异步函数
    !sam.dispatch && (sam.dispatch = function (_type, _payload, _options) {
      var action = utils.unifyObjectStyle(_type, _payload, _options);
      var options = action.options;
      options.async === undefined && (options.async = true);
      return dispatch(action);
    });

    return function (next) {
      // 挂载 commit 方法 至 sam 对象上，用于同步更新状态
      !sam.commit && (sam.commit = function (_type, _payload, _options) {
        // 入参兼容
        return next(utils.unifyObjectStyle(_type, _payload, _options));
      });

      return function (_type, _payload, _options) {
        // 入参兼容
        var action = utils.unifyObjectStyle(_type, _payload, _options);
        var options = action.options;

        // 触发异步方法
        if (options.async) {
          var type = action.type;
          var payload = action.payload;
          var _actions = sam._actions;
          var _actionSubscribers = sam._actionSubscribers;

          var entry = _actions[type];
          if (!entry) {
            if (process.env.NODE_ENV !== 'production') {
              // error(`unknown action type: ${type}`);
              utils.info(("unknown action type: " + type + ", trying mutation type: " + type + " ..."));
            }
            // return;
            entry = [function () {
              return new Promise(function (resolve) {
                resolve(next(action));
              });
            }];
          }

          triggerActionSubscribers(_actionSubscribers, 'before', action, sam.state);

          var result = entry.length > 1
            ? Promise.all(entry.map(function (handler) { return handler(payload); }))
            : entry[0](payload);

          return result.then(function (res) {
            triggerActionSubscribers(_actionSubscribers, 'after', action, sam.state);
            return res;
          });
        }

        // next 为 redux createStore 里面的 dispatch 方法
        return next(action);
      };
    };
  };
}

module.exports = middleware;
