/*
 * redux-sam.js v1.2.0
 * (c) 2018-2019 Jesse Feng
 * Released under the MIT License.
 */
'use strict';

var isString = require('celia/isString');
var isObject = require('celia/isObject');
var forOwn = require('celia/forOwn');

function getLogger(level) {
  /**
   * 输出错误消息
   * @param {...any} args
   */
  return function logger() {
    var args = [], len = arguments.length;
    while ( len-- ) args[ len ] = arguments[ len ];

    console[level].apply(console, [ (((new Date()).toLocaleTimeString()) + " [" + level + "] [redux-sam]") ].concat( args ));
  }
}

var error = getLogger('error');
var info = getLogger('info');

/**
 * 断言
 *
 * @param {Boolean} condition
 * @param {String} msg
 */
function assert(condition, msg) {
  if (!condition) {
    throw new Error(("[redux-sam] " + msg));
  }
}

/**
 * 根据路径获取子state
 *
 * @param {Sam} sam
 * @param {Array} path
 */
function getNestedState(sam, path) {
  var rootState = sam._state;
  return path.length
    ? path.reduce(function (state, key) { return state[key]; }, rootState)
    : rootState;
}

/**
 * 根据路径设置子state
 *
 * @param {Sam} sam
 * @param {Array} path
 * @param {Object} childState
 */
function setNestedState(sam, path, childState) {
  var len = path.length;
  if (len) {
    var lastIndex = len - 1;
    var parentState = getNestedState(sam, path.slice(0, lastIndex));
    var moduleName = path[lastIndex];
    parentState[moduleName] = childState;
  } else {
    sam._state = childState;
  }
  return childState;
}

/**
 * 移除子state
 *
 * @param {Sam} sam
 * @param {Array} path
 */
function removeNestedState(sam, path) {
  var lastIndex = path.length - 1;
  var parentState = getNestedState(sam, path.slice(0, lastIndex));
  var moduleName = path[lastIndex];
  delete parentState[moduleName];
}

/**
 * 统一参数格式为一个对象
 *
 * @param {String|Object} type
 * @param {any} payload
 * @param {Object|undefined} options
 */
function unifyObjectStyle(type, payload, options) {
  var _type;
  if (isObject(type) && (_type = type.type)) {
    options = type.options;
    payload = type.payload;
    type = _type;
  }

  if (process.env.NODE_ENV !== 'production') {
    assert(isString(type), ("expects string as the type, but found " + (typeof type) + "."));
  }

  options = options || {};

  return { type: type, payload: payload, options: options };
}

/**
 * 序列化命名空间
 *
 * @param {Function} fn
 */
function normalizeNamespace(fn) {
  return function (component, map, namespace) {
    isString(namespace)
      ? namespace[namespace.length - 1] !== '/' && (namespace += '/')
      : (namespace = '');
    return fn(component, map, namespace);
  };
}

/**
 * 把数组转换成对象再处理
 *
 * @param {Object|Array} map
 * @param {Function} callback
 */
function normalizeMap(map, callback) {
  return Array.isArray(map)
    ? map.forEach(function (key) { return callback(key, key); })
    : forOwn(map, callback);
}

/**
 * 通过命名空间查找模块
 *
 * @param {Object} sam
 * @param {String} helper
 * @param {String} namespace
 * @return {Object}
 */
function getContextByNamespace(sam, helper, namespace) {
  var ctx = sam._contextNamespaceMap[namespace];
  if (process.env.NODE_ENV !== 'production' && !ctx) {
    error(("module namespace not found in " + helper + "(): " + namespace));
  }
  return ctx;
}

exports.assert = assert;
exports.error = error;
exports.getContextByNamespace = getContextByNamespace;
exports.getNestedState = getNestedState;
exports.info = info;
exports.normalizeMap = normalizeMap;
exports.normalizeNamespace = normalizeNamespace;
exports.removeNestedState = removeNestedState;
exports.setNestedState = setNestedState;
exports.unifyObjectStyle = unifyObjectStyle;
